import math
import re

import poobrains

from poobrains import app, flash, g, new_session, redirect, session
from poobrains.analysis import util

def plot_kind_choices():
    return [(kind, cls.__name__) for (kind, cls) in poobrains.analysis.visualization.Plot.class_children_keyed().items()]

def dtype_choices():
    return [(dtype_name, dtype_name) for dtype_name in util.__dtypes__]

def readable_datasources():

    classes =  {}
    for key, cls in poobrains.analysis.data.EphemeralDataset.class_children_keyed().items():
        try:
            cls.permissions['read'].check(g.user)

        except poobrains.auth.AccessDenied:
            continue # skip EphemeralDataset subclasses the current user is not allowed to read.

        classes[key] = cls

    return classes

def datasource_choices():
    choices = []
    for name, cls in readable_datasources().items():
        if hasattr(cls, 'title') and isinstance(cls.title, str):
            choices.append((cls, cls.title))
        else:
            choices.append((cls, name))

    # TODO: StoredDataset
    return choices


def validate_handle_free(handle):
    if handle in session['apps']['DataEditor']:
        if poobrains.analysis.data.Dataset.select().where(poobrains.analysis.data.Dataset.name == handle).count() == 1:
            raise poobrains.errors.ValidationError(f"Dataset named '{handle}' already exists!")

        else:
            del(session['apps']['DataEditor'][handle]) # FIXME: should technically be in a different function, i guess


class EditorPlotAdd(poobrains.formapp.MultistateFieldset):

    title = "Add plot definition"

    def build(self):

        super().build()

        self['plotdef_name'] = poobrains.form.fields.Text(label='Name', help_text='machine-readable ([a-z], [0-9], -)', validators=[poobrains.form.validators.valid_name])
        self['kind'] = poobrains.form.fields.Select(choices=plot_kind_choices())
        self.plotdef_add = poobrains.form.Button('submit', label="Add plot definition")

    def process(self, submit):

        plot_definition = poobrains.analysis.data.DatasetPlot()
        plot_definition.name = self['plotdef_name'].value
        plot_definition.dataset = self.formapp.instance
        plot_definition.info = {
            'kind': self['kind'].value,
            'layers': {}
        }

        plot_definition.save()

        self.formapp.session['plot_definition_name'] = plot_definition.name
        self.formapp.session['plot_definition_info'] = plot_definition.info
        
        self.parent.tool_active = 'plotdef_options' # show plot options after saving.
        flash(f"Added new plot definition '{plot_definition.name}'.")


class EditorPlotLoad(poobrains.formapp.MultistateFieldset):

    title = "Load another plot definition"

    def build(self):

        super().build()

        if (not self.formapp.plot_definition is None) and \
        self.formapp.plot_definition.info != self.formapp.session['plot_definition_info']:
            self['warning'] = poobrains.form.fields.Message(value="There are unsaved changes for this plot definition!")

        self['plot_definition'] = poobrains.form.fields.Select(label="Plot Definition", type=poobrains.form.types.StorableInstanceParamType(poobrains.analysis.data.DatasetPlot), choices=[(x, x.title) for x in self.formapp.instance.plots])

        self.plotdef_load = poobrains.form.Button('submit', label='Load')

    def validate(self, submit):

        if submit != 'cancel':

            errors = []

            if not self['plot_definition'].value:
                errors.append(poobrains.errors.ValidationError("You need to choose a plot definition.", self['plot_definition']))

            if len(errors):
                raise poobrains.errors.CompoundError(errors)

    def process(self, submit):

        plot_definition = self['plot_definition'].value
        self.formapp.session['plot_definition_name'] = plot_definition.name
        self.formapp.session['plot_definition_info'] = plot_definition.info

        self.parent.reset()
        flash(f"Loaded plot definition '{plot_definition.name}'.")

        return redirect('#_')


class EditorPlotSave(poobrains.formapp.MultistateFieldset):

    def build(self):

        super().build()

        self.title = f"Save plot definition '{self.formapp.session['plot_definition_name']}'"
        self.save = poobrains.form.Button(type='submit', label='Save')

    def process(self, submit):

        plotdef_name = self.formapp.session['plot_definition_name']
        plotdef = poobrains.analysis.data.DatasetPlot.load(plotdef_name)
        plotdef.info = self.formapp.session['plot_definition_info']

        if plotdef.save():
            flash(f"Saved plot definition '{plotdef_name}'.")
            self.session_delete()
            self.parent.tool_active = None
        else:
            flash(f"Failed to save plot definition '{plotdef_name}'!", 'error')

        return redirect('#_')


class EditorPlotKind(poobrains.formapp.MultistateFieldset):

    title = "Change plot kind"

    def build(self):

        super().build()

        self['kind'] = poobrains.form.fields.Select(choices=plot_kind_choices(), default=self.formapp.session['plot_definition_info']['kind'])
        self.apply_kind = poobrains.form.Button('submit', label='Apply')

    def process(self, submit):

        old_plot_kind = self.formapp.session['plot_definition_info']['kind']
        old_plot_cls = poobrains.analysis.visualization.Plot.class_children_keyed()[old_plot_kind]

        new_plot_kind = self['kind'].value
        new_plot_cls = poobrains.analysis.visualization.Plot.class_children_keyed()[new_plot_kind]

        self.formapp.session['plot_definition_info']['kind'] = new_plot_kind

        flash(f"Changed plot kind to {new_plot_kind}.")

        if not old_plot_cls.editor_layer_fieldset is new_plot_cls.editor_layer_fieldset:
            self.formapp.session['plot_definition_info']['layers'].clear()
            flash("Removed old layer information since it's not compatible to the new plot kind.", 'warning')

        self.parent.tool_active = None
        self.session.clear()


class EditorPlotOptions(poobrains.formapp.MultistateFieldset):

    title = "Change plot options"

    def build(self):

        super().build()

        self['plot_options'] = self.formapp.plot_cls.editor_fieldset(self)
        self.apply_options = poobrains.form.Button('submit', label='Apply')

    def process(self, submit):

        self['plot_options'].process(self.child_submit(submit))
        self.parent.reset()
        flash("Applied plot options.")
        return redirect('#_')


class EditorPlotDelete(poobrains.formapp.MultistateFieldset):

    title = "Delete plot definition(s)"

    def build(self):

        super().build()

        self['warning'] = poobrains.form.fields.Message(value="Deletion is permanent!")
        self['plot_definitions'] = poobrains.form.fields.Select(
            type=poobrains.form.types.StorableInstanceParamType(poobrains.analysis.data.DatasetPlot),
            choices=[(plotdef, plotdef.name) for plotdef in self.formapp.instance.plots],
            label='Plot definition(s)',
            multi=True
        )

        self.delete = poobrains.form.Button('submit', label='Delete')

    def process(self, submit):

        current_plotdef_name = self.formapp.session['plot_definition_name']
        current_removed = False
        c = 0
        for plotdef in self['plot_definitions'].value:
            plotdef.delete_instance()
            if plotdef.name == current_plotdef_name:
                current_removed = True
            c += 1

        if current_removed:
            if self.formapp.instance.plots.count():
                plotdef = self.formapp.instance.plots[0]
                self.formapp.session['plot_definition_name'] = plotdef.name 
                self.formapp.session['plot_definition_info'] = plotdef.info

            else:
                self.formapp.session['plot_definition_name'] = None
                self.formapp.session['plot_definition_info'] = None

        self.parent.tool_active = None
        self.session.clear()
        flash(f"Deleted {c} plot definitions.")
        return redirect('#_')


class EditorPlotLayerAdd(poobrains.formapp.MultistateFieldset):

    title = "Add layer"

    def build(self):

        super().build()

        self['layer'] = self.formapp.plot_cls.editor_layer_fieldset(self)
        self.add = poobrains.form.Button('submit', label='Add')

    def process(self, submit):

        self['layer'].process(submit)
        self.parent.tool_active = None
        self.session.clear()


class EditorPlotLayerEdit(poobrains.formapp.MultistateFieldset):

    title = "Edit layer"

    def build(self):

        if self.state == 'start':
            self['layer_name'] = poobrains.form.fields.Select(label="Layer", choices=self.formapp.layer_choices)
            self.edit_choose = poobrains.form.Button('submit', label='Choose')

        elif self.state == 'layer':
            layer_name = self.session['layer_name']
            self.title = f"Edit layer '{layer_name}'"
            self['layer'] = self.formapp.plot_cls.editor_layer_fieldset(self)
            self.edit_apply = poobrains.form.Button('submit', label='Apply')

    def validate(self, submit):

        errors = []

        if submit == 'edit_choose':
            if not self['layer_name'].value:
                errors.append(poobrains.errors.ValidationError("You need to select a layer.", self['layer_name']))
                
        if len(errors):
            raise poobrains.errors.CompoundError(errors)

    def process(self, submit):

        if submit == 'edit_choose':
            self.session.clear()
            self.state = 'layer'
            self.session['layer_name'] = self['layer_name'].value

            flash(f"Now editing layer '{self['layer_name'].value}'.")

        elif submit == 'edit_apply':
            self['layer'].process(self.child_submit(submit))
            self.session_delete()
            self.parent.reset()
            flash("Applied changes.")


class EditorPlotLayerDelete(poobrains.formapp.MultistateFieldset):

    title = "Delete layer(s)"

    def build(self):

        super().build()

        self['layers'] = poobrains.form.fields.Select(label="Layer(s)", choices=self.formapp.layer_choices, multi=True)
        self.delete = poobrains.form.Button('submit', label='Delete')

    def process(self, submit):

        n = 0
        for layer_name in self['layers'].value:
            del(self.formapp.session['plot_definition_info']['layers'][layer_name])
            n += 1

        self.parent.tool_active = None
        self.session.clear()
        flash(f"Deleted {n} layers.")


class EditorPlotToolbox(poobrains.formapp.Toolbox):

    title = "Plot actions"

    tools = {
        'plotdef_load': ('Load plot definition', EditorPlotLoad),
        'plotdef_add': ('Add new plot definition', EditorPlotAdd),
        'plotdef_save': ('Save plot definition', EditorPlotSave),
        'plotdef_kind': ('Change plot kind', EditorPlotKind),
        'plotdef_options': ('Edit plot options', EditorPlotOptions),
        'plotdef_delete': ('Delete plot definition(s)', EditorPlotDelete),
        'layer_add': ('Add layer', EditorPlotLayerAdd),
        'layer_edit': ('Edit layer',EditorPlotLayerEdit),
        'layer_delete': ('Delete layer(s)', EditorPlotLayerDelete)
    }

    def __init__(self, parent, **kwargs):

        super().__init__(parent, **kwargs)

        if self.parent.session['plot_definition_name'] is None: # no associated plotdef exists
            self.tools_enabled = set(('plotdef_load', 'plotdef_add')) # disable all tools that require a loaded plotdef


class EditorDatasetSave(poobrains.formapp.MultistateFieldset):

    title = "Save dataset"
    save = poobrains.form.Button('submit', label='Save')
    save_and_unlock = poobrains.form.Button('submit', label='Save & unlock')

    def process(self, submit):

        if submit == 'save_and_unlock':
            self.formapp.instance.lock = None

        self.formapp.instance.data = self.formapp.dataset.data
        self.formapp.instance.save()
        #self.tags.process(self.child_submit(submit)) # also create TagBindings
        self.parent.tool_active = None
        self.session_delete()
        
        if submit == 'save_and_unlock':
            self.formapp.session_delete()

        flash(f"Saved dataset '{self.formapp.instance.name}'.")
        
        return redirect('#_')


class EditorDatasetMetaEdit(poobrains.formapp.MultistateFieldset):

    title = "Edit base information"

    def build(self):

        super().build()

        self['basedata_form'] = poobrains.form.ProxyFieldset(poobrains.auth.AutoForm(self.formapp.instance, mode='edit'))
        #self['dataset_title'] = poobrains.form.fields.Text(label='Title', default=self.formapp.dataset.title)
        #self['dataset_description'] = poobrains.form.fields.TextArea(label='Description', default=self.formapp.dataset.description)
        self.apply = poobrains.form.Button('submit', label='Apply')

    def process(self, submit):

        if submit == 'apply':
            #self.formapp.dataset.title = self['dataset_title'].value
            #self.formapp.dataset.description = self['dataset_description'].value
            self['basedata_form'].process('save')

            flash("Updated dataset base info.")
            self.parent.tool_active = None
            self.session_delete()

        return redirect('#_')


class EditorObservationEdit(poobrains.formapp.MultistateFieldset):

    def build(self):

        super().build()

        if self.state == 'edit':

            range_bottom = self.session['range_bottom']
            range_top = self.session['range_top']
            if range_bottom == range_top:
                self.title = f"Editing observation #{range_bottom}"
            else:
                self.title = f"Editing observations #{range_bottom} – #{range_top}"

            index_range = range(range_bottom, range_top + 1)

            for idx in index_range:

                observation_form = poobrains.form.Fieldset()
                observation_form.title = f"Observation #{idx}"

                for column_name in self.session['columns']:
                    column = self.formapp.dataset[column_name]
                    column_dtype = util.__dtypes__[column['dtype']]
                    default = column['observations'][idx] if idx in column['observations'] else None
                    observation_form[column_name] = column_dtype.form_field(label=f"{column['title']} ({column['dtype']})", type=column_dtype.form_type, default=default)
                self[f'observation-{idx}'] = observation_form
                self.apply = poobrains.form.Button(type='submit', label='Apply')

        else:

            self.title = 'Edit observation(s)'
            index_range = self.formapp.dataset.complete_index()
            index_min = index_range[0]
            index_max = index_range[-1]

            self['range_bottom'] = poobrains.form.fields.Number(label="First index", min=index_min, max=index_max, default=index_min, type=poobrains.form.types.INT)
            self['range_top'] = poobrains.form.fields.Number(label="Last index", min=index_min, max=index_max, default=index_max, type=poobrains.form.types.INT)
            self['columns'] = poobrains.form.fields.Select(label="Column", choices=self.formapp.dataset.column_choices_of_dtype(list(util.__dtypes__.keys())), multi=True)
            self.edit = poobrains.form.Button(type='submit', label='Edit')

    def process(self, submit):

        if submit == 'edit':
            self.session.clear()
            self.state = 'edit'
            self.session['range_bottom'] = self['range_bottom'].value
            self.session['range_top'] = self['range_top'].value
            self.session['columns'] = self['columns'].value

        else:

            range_bottom = self.session['range_bottom']
            range_top = self.session['range_top']
            index_range = range(range_bottom, range_top + 1)

            for idx in index_range:

                fieldset = self[f'observation-{idx}']
                for column_name in self.session['columns']:
                    self.formapp.dataset[column_name]['observations'][idx] = fieldset[column_name].value

            self.parent.tool_active = None
            self.session.clear()
            flash(f"Updated {range_top - range_bottom} observations.")
        
        return redirect('#_')


class EditorColumnAdd(poobrains.formapp.MultistateFieldset):

    title = "Add new column" 
    
    def build(self):

        super().build()

        self['column_name'] = poobrains.form.fields.Text(label="Name", help_text='machine-readable ([a-z], [0-9], -)', validators=[poobrains.form.validators.valid_name])
        self['column_title'] = poobrains.form.fields.Text(label="Title")
        self['column_dtype'] = poobrains.form.fields.Select(label="Data type", choices=dtype_choices)
        self['column_color_use'] = poobrains.form.fields.Checkbox(label="Use custom color")
        self['column_color'] = poobrains.form.fields.Color(default="#000000")
        self['column_description'] = poobrains.form.fields.TextArea(label="Description")

        self.add = poobrains.form.Button('submit', label='Add')

    def validate(self, submit):

        if submit != 'cancel':

            errors = []

            if self['column_name'].value is None:
                errors.append(poobrains.errors.ValidationError("Need to set a column name.", self['column_name']))

            #if self['column_name'].value in self.formapp.session['dataset']['data']:
            if self['column_name'].value in self.formapp.dataset.keys():
                errors.append(poobrains.errors.ValidationError(f"Column already exists: '{self['column_name'].value}'", self['column_name']))

            if self['column_dtype'].value is None:
                errors.append(poobrains.errors.ValidationError("Need to choose a column data type.", self['column_dtype']))
    
            if len(errors):
                raise poobrains.errors.CompoundError(errors)

    def process(self, submit):

        if self['column_color_use'].value: 
            color = poobrains.svg.color.Color.from_hexrgb(self['column_color'].value)
        else:
            color = None

        self.formapp.dataset[self['column_name'].value] = {
            'title': self['column_title'].value or self['column_name'].value,
            'description': self['column_description'].value,
            'dtype': self['column_dtype'].value,
            'color': color,
            'observations': {}
        }

        self.parent.tool_active = None
        self.session.clear()
        flash(f"Created column '{self['column_name'].value}'.")

        return redirect('#_')


class EditorDatasetMerge(poobrains.formapp.MultistateFieldset):

    """
    States:

    * `start`:              Presents source choice.
    * `source_parameters`:  Presents source-specific fieldset to instantiate the
                            `other` dataset.
    * `merge`:              Presents merge options for the instantiated `other`
                            dataset, as well as the option to add a new join
                            condition.
    * `join_self`:          Presents the choice for the column on the edited
                            dataset to use for a new join condition.
    * `join_other`:         Presents the choice for the column on the `other`
                            dataset, as well as dtype-specific tolerance
                            options.
    """

    title = "Merge dataset"

    def build(self):

        super().build()

        self['source'] = poobrains.form.fields.Select(choices=datasource_choices, type=poobrains.form.types.StorableParamType(poobrains.analysis.EphemeralDataset))

        if 'source' in self.session:
                self['source'].value = readable_datasources()[self.session['source']]

        if self.state != 'start':
            self['source'].readonly = True

        if self.state == 'source_parameters':
            self['source_parameters'] = self.summon_parameter_fieldset(self['source'].value)

        elif self.state == 'merge':

            self.other = poobrains.analysis.data.EphemeralDataset.from_dict(self.session['other'], whole=True)
            self['other_preview'] = poobrains.form.fields.RenderableWrapper(self.other.table())

            self['other_prefix'] = poobrains.form.fields.Text(label='Other prefix', help_text='Prefix to be applied to column names of the dataset being merged into the current one.')

            if len(self.session['joins']):
                for idx, joindef in enumerate(self.session['joins']):
                    fieldset = poobrains.form.Fieldset(title=f"Join condition #{idx}")
                    fieldset.message = poobrains.form.fields.Message(value=joindef['description'])
                    fieldset.delete = poobrains.form.Button('submit', label='Delete')
                    self[f"join-{idx}"] = fieldset


            self.join_add = poobrains.form.Button('submit', label='Add join')
            self.merge = poobrains.form.Button('submit', label='Merge')

        elif self.state in ('join_self', 'join_other'):

            self['join_column_self'] = poobrains.form.fields.Select(
                label="Column (self)", 
                choices=self.formapp.dataset.column_choices,
                help_text="Column of the *edited* dataset that is used to identify matching rows by."
            )

            if self.state == 'join_self':

                self.choose = poobrains.form.Button('submit', label='Choose')

            else: # join_other
            
                self.other = poobrains.analysis.data.EphemeralDataset.from_dict(self.session['other'], whole=True)

                self['join_column_self'].value = self.session['join_column_self']
                self['join_column_self'].readonly = True

                column_self = self.formapp.dataset[self.session['join_column_self']]
                dtype = util.__dtypes__[column_self['dtype']]

                self['join_dtype_form'] = dtype.join_form(self)

        if self.state in ('start', 'source_parameters'):
            self.load = poobrains.form.Button('submit', label='Load')

    def summon_parameter_fieldset(self, source):

        if issubclass(source, poobrains.analysis.sources.CustomFormDataset):
            return source.editor_fieldset(self)

        else:
            subform = poobrains.form.Fieldset()
            for param_name, param_info in poobrains.helpers.function_parameters(source.load).items():
                param_type = poobrains.form.types.lookup_table[param_info['type']]()
                subform[param_name] = poobrains.form.fields.Text(label=param_name, type=param_type, default=param_info['default'])

        return subform

    def validate(self, submit):

        if submit != 'cancel':

            errors = []
            if 'source' in self and self['source'].value is None:
                errors.append(poobrains.errors.ValidationError("Need to choose a data source", self['source']))

            if len(errors):
                raise poobrains.errors.CompoundError(errors)

    def process(self, submit):

        if self.state == 'start':

            source = self['source'].value

            source_parameters = poobrains.helpers.function_parameters(source.load)

            self.session.clear()
            self.session['source'] = source.__name__
            self.session['joins'] = []

            if len(source_parameters) == 0:
                ds = source.load()
                self.state = 'merge'
                self.session['other'] = ds.to_dict(whole=True)

            else:
                self.state = 'source_parameters'

        elif self.state == 'source_parameters':

            source = self['source'].value

            if issubclass(source, poobrains.analysis.sources.CustomFormDataset):
                self['source_parameters'].process(self.child_submit(submit)) # will set self.session['other'], if applicable

            else:

                source_parameters = poobrains.helpers.function_parameters(source.load)

                parameters = {}
                for param_name in source_parameters:
                    if param_name in self['source_parameters']:
                        parameters[param_name] = self['source_parameters'][param_name].value

                if len(parameters) == len(source_parameters):
                    other = source.load(**parameters)

                    self.state = 'merge'
                    self.session['other'] = other.to_dict(whole=True)

                else:
                    flash("Internal bleeding detected.", 'error')

        elif self.state == 'merge':

            if submit == 'join_add':
                self.state = 'join_self'
                return redirect('#_')

            elif submit == 'merge':

                try:

                    self.formapp.dataset.merge(self.other, joins=self.session['joins'], operation='and', prefix=self['other_prefix'].value)

                except Exception as e: # FIXME: too broad?
                    flash(f"Error when merging: {str(e)}", 'error')
                else:

                    self.parent.tool_active = None
                    self.session.clear()

                    flash("Merge successful.")

            else: # should™ be delete button of a join condition
                match = re.match('^join-(\d*)\.delete$', submit)
                if match:
                    idx = int(match.groups()[0])
                    self.session['joins'].pop(idx)
                    flash(f"Removed join condition #{idx}")

            return redirect('#_')

        elif self.state == 'join_self':

            self.session['join_column_self'] = self['join_column_self'].value
            self.state = 'join_other'

        elif self.state == 'join_other':
            self['join_dtype_form'].process(self.child_submit(submit)) 


class EditorColumnMetaEdit(poobrains.formapp.MultistateFieldset):

    title = "Edit column information"

    def build(self):

        super().build()

        if self.state == 'column':

            column_name = self.session['column_name']
            column = self.formapp.dataset[column_name]

            color_value = None if column['color'] is None else column['color'].hexrgb()

            self.title = f"Edit column information for '{column_name}'"

            self['column_name'] = poobrains.form.fields.Text(label='Name', value=column_name, help_text='machine readable ([a-z], [0-9], -)')
            self['column_title'] = poobrains.form.fields.Text(label='Title', value=column['title'])
            self['column_color_use'] = poobrains.form.fields.Checkbox(label="Use custom color", value=not column['color'] is None)
            self['column_color'] = poobrains.form.fields.Color(label='Color', value=color_value)
            self['column_description'] = poobrains.form.fields.TextArea(label='Description', value=column['description'], help_text='Describe the axes, summarize the data and any notable trends in it.')

            self.apply = poobrains.form.Button('submit', label='Apply')

        else:

            self['column'] = poobrains.form.fields.Select(label='Column', choices=self.formapp.dataset.column_choices)

            if 'column_name' in self.session:
                self['column'].value = self.session['column_name']

            self.column_choose = poobrains.form.Button('submit', label='Choose')

    def process(self, submit):

        if submit == 'column_choose':

            self.session.clear()
            self.state = 'column'
            self.session['column_name'] = self['column'].value
            flash(f"Now editing column {self['column'].value}.")

        elif submit == 'apply':

            column_name = self.session['column_name']
            authoritative_column_name = self['column_name'].value

            if column_name != authoritative_column_name: # means this column is getting renamed.
                self.formapp.dataset[authoritative_column_name] = self.formapp.dataset[column_name]
                del(self.formapp.dataset[column_name])

                if self.formapp.dataset.plottable:
                    # rename any occurence of this renamed column in plot definitions
                    for plot_definition in self.formapp.dataset.plots:

                        plot_kind = plot_definition.info['kind']
                        plot_cls = poobrains.analysis.visualization.Plot.class_children_keyed()[plot_kind]

                        for layer_name, layer_info in plot_definition.info['layers'].items():
                            for key in plot_cls.layer_column_fields:
                                if layer_info[key] == column_name:
                                    layer_info[key] = authoritative_column_name

                flash(f"Renamed column from '{column_name}' to '{authoritative_column_name}'.")

            self.formapp.dataset[authoritative_column_name]['title'] = self['column_title'].value
            self.formapp.dataset[authoritative_column_name]['description'] = self['column_description'].value
            
            if self['column_color_use'].value:
                self.formapp.dataset[authoritative_column_name]['color'] = poobrains.svg.color.Color.from_hexrgb(self['column_color'].value)
            else:
                self.formapp.dataset[authoritative_column_name]['color'] = None

            self.parent.tool_active = None
            self.session.clear()
            flash(f"Updated information for column '{authoritative_column_name}'.")

        return redirect('#_')


class EditorColumnDelete(poobrains.formapp.MultistateFieldset):

    title = "Delete column(s)"

    def build(self):
        super().build()
        
        self['columns'] = poobrains.form.fields.Select(label="Column", choices=self.formapp.dataset.column_choices, multi=True)
        self.delete = poobrains.form.Button('submit', label='Delete')

    def process(self, submit):

        if submit == 'delete':

            i = 0
            for column_name in self['columns'].value:
                del(self.formapp.dataset[column_name])
                i += 1

            self.parent.tool_active = None
            self.session.clear()
            if i == 1:
                flash(f"Deleted column '{column_name}'.")
            else:
                flash(f"Deleted {i} columns.")

        else:
            flash(f"{self.name} did nothing. This might be a bug.", 'warning')

        return redirect('#_')


class EditorFilter(poobrains.formapp.MultistateFieldset):

    title = "Filter"

    def build(self):

        super().build()

        if 'filters' in self.session:

            i = 0
            for filterdef in self.session['filters']:

                fieldset = poobrains.form.Fieldset(title=f"Filter #{i}")
                fieldset.message = poobrains.form.fields.Message(value=filterdef['description'])
                fieldset.delete = poobrains.form.Button('submit', label='Delete')
                self[f"filter-{i}"] = fieldset
                i += 1

        if self.state == 'start':
            #self.filter_mode = poobrains.form.fields.Select(label='Mode', help_text='Whether to include or exclude matched rows', choices=(('include', 'include'), ('exclude', 'exclude')))
            self.filter_operation = poobrains.form.fields.Select(label='Operator', help_text='Boolean operation used between filters when applying', choices=(('and', 'AND'),('or', 'OR')))
            self.filter_new = poobrains.form.Button('submit', label='Add filter')
            self.apply = poobrains.form.Button('submit', label='Apply')

        elif self.state in ('filter_column', 'filter_op'):
            self.filter_params = poobrains.form.Fieldset(title="Add filter")
            self.filter_params['column'] = poobrains.form.fields.Select(label='Column', choices=self.formapp.dataset.column_choices)

            if self.state == 'filter_column':
                self.filter_params.filter_column = poobrains.form.Button('submit', label='Continue')

            if self.state == 'filter_op':

                column_name = self.session['filter_column']
                dtype = util.__dtypes__[self.formapp.dataset[column_name]['dtype']]

                self.filter_params['column'].value = column_name
                self.filter_params['column'].readonly = True
                
                self.filter_params['filter_dtype_form'] = dtype.filter_form(self)

    def validate(self, submit):

        errors = []

#        if not 'filters' in self.session or len(self.session['filters']) == 0:
#
#            if self.filter_params['column'].value is None:
#                errors.append(poobrains.errors.ValidationError('You must select a column.', self.filter_params['column']))
#
#            if self.filter_params['op'].value is None:
#                errors.append(poobrains.errors.ValidationError('You must select an operation.', self.filter_params['op']))
#
#            if self.filter_params['value'].value is None and self.filter_params['value_column'].value is None:
#                e = poobrains.errors.ValidationError('You must select a value to compare against.')
#                if self.filter_params['value'].value is None:
#                    self.filter_params['value'].errors.append(e)
#                if self.filter_params['value_column'].value is None:
#                    self.filter_params['value_column'].errors.append(e)
#                errors.append(e)

        if len(errors) > 0:
            raise poobrains.errors.CompoundError(errors)

    def process(self, submit):

        if submit == 'filter_new':
            self.state = 'filter_column'

        elif submit == 'filter_params.filter_column':

            self.state = 'filter_op'
            self.session['filter_column'] = self['filter_params']['column'].value

        elif submit.startswith('filter_params.filter_dtype_form.'):

            if not 'filters' in self.session:
                self.session['filters'] = []

            self.process_submitting_fieldset(submit)

        elif submit == 'apply':

            num_rows_old = self.formapp.dataset.rowcount

            filterdefs = self.session.get('filters', [])
            self.formapp.dataset.filter(filterdefs, operation=self['filter_operation'].value)

            num_filters = len(filterdefs)
            num_rows_new = self.formapp.dataset.rowcount

            self.parent.tool_active = None
            self.session.clear()

            flash(f"Applied {num_filters} filters. {num_rows_new} remaining after removing {num_rows_old - num_rows_new} rows.")

        else:
            match = re.match('^filter-(\d*)\.delete$', submit)
            if match:
                idx = int(match.groups()[0])
                self.session['filters'].pop(idx)
                flash(f"Removed filter #{idx}")

        return redirect('#_')


class EditorDataToolbox(poobrains.formapp.Toolbox):

    title = "Actions"

    tools = {
        'dataset_save': ('Save dataset', EditorDatasetSave),
        'dataset_meta_edit': ('Edit dataset info', EditorDatasetMetaEdit),
        'dataset_merge': ('Merge dataset', EditorDatasetMerge),
        'filter': ('Filter', EditorFilter),
        'observation_edit': ('Edit observation(s)', EditorObservationEdit),
        'column_add': ('Add column', EditorColumnAdd),
        'column_meta_edit': ('Edit column metadata', EditorColumnMetaEdit),
        'column_delete': ('Delete column(s)', EditorColumnDelete)
    }


class DataEditor(poobrains.formapp.BoundFormApp):

    def __init__(self, instance, mode='add'):

        super().__init__(instance, mode=mode)

        self.title = f"Edit dataset '{self.instance.name}'."

        if self.session_is_new:

            self.session['tab_form'] = 'data_toolbox'

            self.dataset = self.instance
            self.session['dataset'] = self.dataset.to_dict(whole=True)
        
            if self.instance.plots.count() > 0:
                plot_def = self.instance.plots[0]
                self.plot_definition = plot_def
                self.session['plot_definition_name'] = plot_def.name
                self.session['plot_definition_info'] = plot_def.info
            else:
                self.plot_definition = None
                self.session['plot_definition_name'] = None
                self.session['plot_definition_info'] = None

        else:

            self.dataset = poobrains.analysis.data.EphemeralDataset.from_dict(self.session['dataset'], whole=True)

            if self.session['plot_definition_name'] is None: 
                self.plot_definition = None
            else:
                self.plot_definition = poobrains.analysis.data.DatasetPlot.load(self.session['plot_definition_name']) # TODO: don't we want to use EphemeralDatasetPlot with session['plot_definition_info']?

        self['display'] = self.display_tab()

        plot_toolbox = EditorPlotToolbox(self, name='plot_toolbox', css_class='action-control')
        data_toolbox = EditorDataToolbox(self, name='data_toolbox', css_class='action-control')

        self['form'] = poobrains.form.TabbedFieldsets([plot_toolbox, data_toolbox])
        self['form'].value = self.session['tab_form']

        if self.session_is_new:
            self.session['tab_display'] = self['display'].value
            self.session['tab_form'] = self['form'].value
        else:
            self['display'].value = self.session['tab_display']
            self['form'].value = self.session['tab_form']

    @property
    def plot_kind(self):
        if self.session['plot_definition_info'] is None:
            return None
        return self.session['plot_definition_info']['kind']

    @property
    def plot_cls(self):
        if not self.plot_kind is None:
            return poobrains.analysis.visualization.Plot.class_children_keyed()[self.plot_kind]

    @property
    def layer_choices(self):
        return [(layer_name, layer_name) for layer_name in self.session['plot_definition_info']['layers'].keys()]

    @property
    def is_locked(self):
        return self.instance.is_locked

    @property
    def has_lock(self):
        return self.instance.has_lock

    def display_tab(self):

        if isinstance(self.session['plot_definition_info'], dict):
            plot = self.dataset.plot(self.session['plot_definition_info'])
        else:
            plot = poobrains.rendering.RenderString("No plot definition.", name='Plot')

        table = self.dataset.table(max_rows=math.inf)

        plot_fieldset = poobrains.form.Fieldset(name="plot-wrapper")

        table_fieldset = poobrains.form.Fieldset(name="table-wrapper")

        tabs = poobrains.form.TabbedFieldsets([plot_fieldset, table_fieldset])
        table_fieldset['table'] = poobrains.form.fields.RenderableWrapper(table)
        plot_fieldset['plot'] = poobrains.form.fields.RenderableWrapper(plot)
        table_fieldset.title = table.title
        plot_fieldset.title = plot.title

        return tabs

    def validate(self, submit):

        if '.' in submit:
            self.validate_submitting_fieldset(submit)

    def process(self, submit):

        self.session['tab_display'] = self['display'].value
        self.session['tab_form'] = self['form'].value

        r = self.process_submitting_fieldset(submit)

        if self.has_lock:
            self.session['dataset'] = self.dataset.to_dict(whole=True) # write the current state of the dataset into session in case it was modified

        if r is None:
            return self
        return r
